#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
import unittest

from rdflib import Graph, Literal, URIRef

from oc_ocdm.graph.graph_set import GraphSet
from oc_ocdm.prov.prov_set import ProvSet
from oc_ocdm.prov.entities.snapshot_entity import SnapshotEntity

from oc_ocdm.counter_handler.sqlite_counter_handler import SqliteCounterHandler


class TestProvSet(unittest.TestCase):
    resp_agent = 'http://resp_agent.test/'

    def setUp(self):
        self.graph_set = GraphSet("http://test/", "./info_dir/", "", False)
        self.prov_set = ProvSet(self.graph_set, "http://test/", "./info_dir/", False, custom_counters={'ci': SqliteCounterHandler('oc_ocdm/test/prov/prov_counter.db')})

    def test_add_se(self):
        prov_subj = self.graph_set.add_br(self.resp_agent)
        se = self.prov_set.add_se(prov_subj)

        self.assertIsNotNone(se)
        self.assertIsInstance(se, SnapshotEntity)
        self.assertEqual(str(se.g.identifier), str(prov_subj.res) + "/prov/")

    def test_generate_provenance(self):
        cur_time = 1607375859.846196
        cur_time_str = '2020-12-07T21:17:39+00:00'

        with self.subTest('Creation [Merged entity]'):
            a = self.graph_set.add_br(self.resp_agent)
            b = self.graph_set.add_br(self.resp_agent)
            a.merge(b)

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            se_a = self.prov_set.get_entity(URIRef(a.res + '/prov/se/1'))
            self.assertIsNotNone(se_a)
            self.assertIsInstance(se_a, SnapshotEntity)
            self.assertEqual(a.res, se_a.get_is_snapshot_of())
            self.assertEqual(cur_time_str, se_a.get_generation_time())
            if a.source is not None:
                self.assertEqual(a.source, str(se_a.get_primary_source()))
            if a.resp_agent is not None:
                self.assertEqual(a.resp_agent, str(se_a.get_resp_agent()))

            self.assertEqual(f"The entity '{a.res}' has been created.", se_a.get_description())
        with self.subTest('No snapshot [Merged entity]'):
            a = self.graph_set.add_br(self.resp_agent)
            b = self.graph_set.add_br(self.resp_agent)
            a.merge(b)
            se_a_1 = self.prov_set.add_se(a)

            # This avoids that the presence of the mandatory rdf:type gets interpreted
            # as a modification with respect to an empty preexisting_graph:
            a.remove_every_triple()

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            se_a_2 = self.prov_set.get_entity(URIRef(a.res + '/prov/se/2'))
            self.assertIsNone(se_a_2)
        with self.subTest('Modification [Merged entity]'):
            title = "TEST TITLE"
            a = self.graph_set.add_br(self.resp_agent)
            b = self.graph_set.add_br(self.resp_agent)
            b.has_title(title)
            a.merge(b)
            se_a_1 = self.prov_set.add_se(a)

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)
            self.assertEqual(cur_time_str, se_a_1.get_invalidation_time())

            se_a_2 = self.prov_set.get_entity(URIRef(a.res + '/prov/se/2'))
            self.assertIsNotNone(se_a_2)
            self.assertIsInstance(se_a_2, SnapshotEntity)
            self.assertEqual(a.res, se_a_2.get_is_snapshot_of())


            self.assertEqual(cur_time_str, se_a_2.get_generation_time())
            if a.source is not None:
                self.assertEqual(a.source, str(se_a_2.get_primary_source()))
            if a.resp_agent is not None:
                self.assertEqual(a.resp_agent, str(se_a_2.get_resp_agent()))

            self.assertSetEqual({se_a_1}, set(se_a_2.get_derives_from()))
            self.assertIsNotNone(se_a_2.get_update_action())
            self.assertEqual(f"The entity '{a.res}' has been modified.", se_a_2.get_description())
        with self.subTest('Merge [Merged entity]'):
            a = self.graph_set.add_br(self.resp_agent)
            b = self.graph_set.add_br(self.resp_agent)
            c = self.graph_set.add_br(self.resp_agent)
            a.merge(b)
            a.merge(c)
            se_a_1 = self.prov_set.add_se(a)
            se_b_1 = self.prov_set.add_se(b)

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            self.assertEqual(cur_time_str, se_a_1.get_invalidation_time())

            se_a_2 = self.prov_set.get_entity(URIRef(a.res + '/prov/se/2'))
            self.assertIsNotNone(se_a_2)
            self.assertIsInstance(se_a_2, SnapshotEntity)
            self.assertEqual(a.res, se_a_2.get_is_snapshot_of())
            self.assertEqual(cur_time_str, se_a_2.get_generation_time())
            if a.source is not None:
                self.assertEqual(a.source, str(se_a_2.get_primary_source()))
            if a.resp_agent is not None:
                self.assertEqual(a.resp_agent, str(se_a_2.get_resp_agent()))

            self.assertSetEqual({se_a_1, se_b_1}, set(se_a_2.get_derives_from()))
            self.assertEqual(f"The entity '{a.res}' has been merged with '{b.res}'.", se_a_2.get_description())
        with self.subTest('Creation [Non-Merged entity]'):
            a = self.graph_set.add_br(self.resp_agent)

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            se_a = self.prov_set.get_entity(URIRef(a.res + '/prov/se/1'))
            self.assertIsNotNone(se_a)
            self.assertIsInstance(se_a, SnapshotEntity)
            self.assertEqual(a.res, se_a.get_is_snapshot_of())
            self.assertEqual(cur_time_str, se_a.get_generation_time())
            if a.source is not None:
                self.assertEqual(a.source, str(se_a.get_primary_source()))
            if a.resp_agent is not None:
                self.assertEqual(a.resp_agent, str(se_a.get_resp_agent()))

            self.assertEqual(f"The entity '{a.res}' has been created.", se_a.get_description())
        with self.subTest('No snapshot [Non-Merged entity] (Scenario 1)'):
            a = self.graph_set.add_br(self.resp_agent)
            a.mark_as_to_be_deleted()

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            se_a = self.prov_set.get_entity(URIRef(a.res + '/prov/se/1'))
            self.assertIsNone(se_a)
        with self.subTest('No snapshot [Merged entity] (Scenario 2)'):
            a = self.graph_set.add_br(self.resp_agent)
            se_a_1 = self.prov_set.add_se(a)

            # This avoids that the presence of the mandatory rdf:type gets interpreted
            # as a modification with respect to an empty preexisting_graph:
            a.remove_every_triple()

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            se_a_2 = self.prov_set.get_entity(URIRef(a.res + '/prov/se/2'))
            self.assertIsNone(se_a_2)
        with self.subTest('Deletion [Non-Merged entity]'):
            title = "TEST TITLE"
            a = self.graph_set.add_br(self.resp_agent)
            se_a_1 = self.prov_set.add_se(a)
            a.has_title(title)
            a.mark_as_to_be_deleted()

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            self.assertEqual(cur_time_str, se_a_1.get_invalidation_time())

            se_a_2 = self.prov_set.get_entity(URIRef(a.res + '/prov/se/2'))
            self.assertIsNotNone(se_a_2)
            self.assertIsInstance(se_a_2, SnapshotEntity)
            self.assertEqual(a.res, se_a_2.get_is_snapshot_of())
            self.assertEqual(cur_time_str, se_a_2.get_generation_time())
            if a.source is not None:
                self.assertEqual(a.source, str(se_a_2.get_primary_source()))
            if a.resp_agent is not None:
                self.assertEqual(a.resp_agent, str(se_a_2.get_resp_agent()))

            self.assertSetEqual({se_a_1}, set(se_a_2.get_derives_from()))
            self.assertEqual(f"The entity '{a.res}' has been deleted.", se_a_2.get_description())
        with self.subTest('Modification [Non-Merged entity]'):
            title = "TEST TITLE"
            a = self.graph_set.add_br(self.resp_agent)
            se_a_1 = self.prov_set.add_se(a)
            a.has_title(title)

            result = self.prov_set.generate_provenance(cur_time)
            self.assertIsNone(result)

            self.assertEqual(cur_time_str, se_a_1.get_invalidation_time())

            se_a_2 = self.prov_set.get_entity(URIRef(a.res + '/prov/se/2'))
            self.assertIsNotNone(se_a_2)
            self.assertIsInstance(se_a_2, SnapshotEntity)
            self.assertEqual(a.res, se_a_2.get_is_snapshot_of())
            self.assertEqual(cur_time_str, se_a_2.get_generation_time())
            if a.source is not None:
                self.assertEqual(a.source, str(se_a_2.get_primary_source()))
            if a.resp_agent is not None:
                self.assertEqual(a.resp_agent, str(se_a_2.get_resp_agent()))

            self.assertSetEqual({se_a_1}, set(se_a_2.get_derives_from()))
            self.assertIsNotNone(se_a_2.get_update_action())
            self.assertEqual(f"The entity '{a.res}' has been modified.", se_a_2.get_description())

    def test_retrieve_last_snapshot(self):
        br = self.graph_set.add_br(self.resp_agent)
        br_res = br.res
        result = self.prov_set._retrieve_last_snapshot(br_res)
        self.assertIsNone(result)

        se = self.prov_set.add_se(br)
        result = self.prov_set._retrieve_last_snapshot(br_res)
        self.assertIsNotNone(result)
        self.assertEqual(str(result), str(se))

        prov_subject = URIRef('https://w3id.org/oc/corpus/br/0')
        self.assertRaises(ValueError, self.prov_set._retrieve_last_snapshot, prov_subject)

        prov_subject = URIRef('https://w3id.org/oc/corpus/br/-1')
        self.assertRaises(ValueError, self.prov_set._retrieve_last_snapshot, prov_subject)

        prov_subject = URIRef('https://w3id.org/oc/corpus/br/abc')
        self.assertRaises(ValueError, self.prov_set._retrieve_last_snapshot, prov_subject)

    def test_generate_provenance_for_citations(self):
        preexisting_graph = Graph()
        preexisting_graph.add((
            URIRef('https://w3id.org/oc/index/coci/ci/020010000023601000907630001040258020000010008010559090238044008040338381018136312231227010309014203370037122439026325-020010305093619112227370109090937010437073701020309'),
            URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'),
            URIRef('http://purl.org/spar/cito/Citation')))
        preexisting_graph.add((
            URIRef('https://w3id.org/oc/index/coci/ci/020010000023601000907630001040258020000010008010559090238044008040338381018136312231227010309014203370037122439026325-020010305093619112227370109090937010437073701020309'),
            URIRef('http://purl.org/spar/cito/hasCitationCreationDate'),
            Literal('2022', datatype='http://www.w3.org/2001/XMLSchema#gYear')))
        ci = self.graph_set.add_ci(self.resp_agent, res=URIRef('https://w3id.org/oc/index/coci/ci/020010000023601000907630001040258020000010008010559090238044008040338381018136312231227010309014203370037122439026325-020010305093619112227370109090937010437073701020309'), preexisting_graph=preexisting_graph)
        self.prov_set.generate_provenance()
        self.graph_set.commit_changes()
        ci.has_citation_creation_date('2022')
        self.prov_set.generate_provenance()
        prov_entity = self.prov_set._retrieve_last_snapshot(URIRef('https://w3id.org/oc/index/coci/ci/020010000023601000907630001040258020000010008010559090238044008040338381018136312231227010309014203370037122439026325-020010305093619112227370109090937010437073701020309'))
        self.assertEqual(prov_entity, URIRef('https://w3id.org/oc/index/coci/ci/020010000023601000907630001040258020000010008010559090238044008040338381018136312231227010309014203370037122439026325-020010305093619112227370109090937010437073701020309/prov/se/1'))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2023, Arcangelo Massari <arcangelo.massari@unibo.it>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.

import sqlite3

from counter_handler.counter_handler import CounterHandler


class SqliteCounterHandler(CounterHandler):
    """A concrete implementation of the ``CounterHandler`` interface that persistently stores
    the counter values within a SQLite database."""

    def __init__(self, database: str) -> None:
        """
        Constructor of the ``SqliteCounterHandler`` class.

        :param database: The name of the database
        :type info_dir: str
        """
        sqlite3.threadsafety = 3
        self.con = sqlite3.connect(database)
        self.cur = self.con.cursor()
        self.cur.execute("""CREATE TABLE IF NOT EXISTS info(
            entity TEXT PRIMARY KEY, 
            count INTEGER)""")

    def set_counter(self, new_value: int, entity_name: str) -> None:
        """
        It allows to set the counter value of provenance entities.

        :param new_value: The new counter value to be set
        :type new_value: int
        :param entity_name: The entity name
        :type entity_name: str
        :raises ValueError: if ``new_value`` is a negative integer.
        :return: None
        """
        entity_name = str(entity_name)
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")
        self.cur.execute(f"INSERT OR REPLACE INTO info (entity, count) VALUES ('{entity_name}', {new_value})")
        self.con.commit()

    def read_counter(self, entity_name: str) -> int:
        """
        It allows to read the counter value of provenance entities.

        :param entity_name: The entity name
        :type entity_name: str
        :return: The requested counter value.
        """
        entity_name = str(entity_name)
        result = self.cur.execute(f"SELECT count FROM info WHERE entity='{entity_name}'")
        rows = result.fetchall()
        if len(rows) == 1:
            return rows[0][0]
        elif len(rows) == 0:
            return 0
        else:
            raise(Exception("There is more than one counter for this entity. The databse id broken"))

    def increment_counter(self, entity_name: str) -> int:
        """
        It allows to increment the counter value of graph and provenance entities by one unit.

        :param entity_name: The entity name
        :type entity_name: str
        :return: The newly-updated (already incremented) counter value.
        """
        entity_name = str(entity_name)
        cur_count = self.read_counter(entity_name)
        count = cur_count + 1
        self.set_counter(count, entity_name)
        return count

    def increment_metadata_counter(self):
        pass
    
    def read_metadata_counter(self):
        pass
    
    def set_metadata_counter(self):
        pass


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only
from oc_ocdm.support.support import get_datatype_from_iso_8601

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.bibliographic_reference import BibliographicReference
    from oc_ocdm.graph.entities.bibliographic.agent_role import AgentRole
    from oc_ocdm.graph.entities.bibliographic.discourse_element import DiscourseElement
    from oc_ocdm.graph.entities.bibliographic.resource_embodiment import ResourceEmbodiment
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class BibliographicResource(BibliographicEntity):
    """Bibliographic resource (short: br): a published bibliographic resource that cites/is
       cited by another published bibliographic resource."""

    @accepts_only('br')
    def merge(self, other: BibliographicResource) -> None:
        """
        The merge operation allows combining two ``BibliographicResource`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``BibliographicResource``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(BibliographicResource, self).merge(other)

        title: Optional[str] = other.get_title()
        if title is not None:
            self.has_title(title)

        subtitle: Optional[str] = other.get_subtitle()
        if subtitle is not None:
            self.has_subtitle(subtitle)

        container: Optional[BibliographicResource] = other.get_is_part_of()
        if container is not None:
            self.is_part_of(container)

        citations_list: List[BibliographicResource] = other.get_citations()
        for cur_citation in citations_list:
            self.has_citation(cur_citation)

        pub_date: Optional[str] = other.get_pub_date()
        if pub_date is not None:
            self.has_pub_date(pub_date)

        re_list: List[ResourceEmbodiment] = other.get_formats()
        for cur_format in re_list:
            self.has_format(cur_format)

        number: Optional[str] = other.get_number()
        if number is not None:
            self.has_number(number)

        edition: Optional[str] = other.get_edition()
        if edition is not None:
            self.has_edition(edition)

        be_list: List[BibliographicReference] = other.get_contained_in_reference_lists()
        for reference in be_list:
            self.contains_in_reference_list(reference)

        de_list: List[DiscourseElement] = other.get_contained_discourse_elements()
        for discourse_element in de_list:
            self.contains_discourse_element(discourse_element)

        ar_list: List[AgentRole] = other.get_contributors()
        for agent_role in ar_list:
            self.has_contributor(agent_role)

        related_doc_list: List[URIRef] = other.get_related_documents()
        for doc in related_doc_list:
            self.has_related_document(doc)

    # HAS TITLE
    def get_title(self) -> Optional[str]:
        """
        Getter method corresponding to the ``dcterms:title`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_title)

    @accepts_only('literal')
    def has_title(self, string: str) -> None:
        """
        Setter method corresponding to the ``dcterms:title`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The title of the bibliographic resource.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_title()
        self._create_literal(GraphEntity.iri_title, string)

    def remove_title(self) -> None:
        """
        Remover method corresponding to the ``dcterms:title`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_title, None))

    # HAS SUBTITLE
    def get_subtitle(self) -> Optional[str]:
        """
        Getter method corresponding to the ``fabio:hasSubtitle`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_subtitle)

    @accepts_only('literal')
    def has_subtitle(self, string: str) -> None:
        """
        Setter method corresponding to the ``fabio:hasSubtitle`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The subtitle of the bibliographic resource.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_subtitle()
        self._create_literal(GraphEntity.iri_has_subtitle, string)

    def remove_subtitle(self) -> None:
        """
        Remover method corresponding to the ``fabio:hasSubtitle`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_subtitle, None))

    # IS PART OF (BibliographicResource)
    def get_is_part_of(self) -> Optional[BibliographicResource]:
        """
        Getter method corresponding to the ``frbr:partOf`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_part_of, 'br')
        if uri is not None:
            return self.g_set.add_br(self.resp_agent, self.source, uri)

    @accepts_only('br')
    def is_part_of(self, br_res: BibliographicResource) -> None:
        """
        Setter method corresponding to the ``frbr:partOf`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The corpus identifier of the bibliographic resource (e.g. issue, volume, journal,
        conference proceedings) that contains the subject bibliographic resource.`

        :param br_res: The value that will be set as the object of the property related to this method
        :type br_res: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_is_part_of()
        self.g.add((self.res, GraphEntity.iri_part_of, br_res.res))

    def remove_is_part_of(self) -> None:
        """
        Remover method corresponding to the ``frbr:partOf`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_part_of, None))

    # CITES (BibliographicResource)
    def get_citations(self) -> List[BibliographicResource]:
        """
        Getter method corresponding to the ``cito:cites`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_cites, 'br')
        result: List[BibliographicResource] = []
        for uri in uri_list:
            result.append(self.g_set.add_br(self.resp_agent, self.source, uri))
        return result

    @accepts_only('br')
    def has_citation(self, br_res: BibliographicResource) -> None:
        """
        Setter method corresponding to the ``cito:cites`` RDF predicate.

        `The corpus identifier of the bibliographic resource cited by the subject bibliographic
        resource.`

        :param br_res: The value that will be set as the object of the property related to this method
        :type br_res: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_cites, br_res.res))

    @accepts_only('br')
    def remove_citation(self, br_res: BibliographicResource = None) -> None:
        """
        Remover method corresponding to the ``cito:cites`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param br_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type br_res: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if br_res is not None:
            self.g.remove((self.res, GraphEntity.iri_cites, br_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_cites, None))

    # HAS PUBLICATION DATE
    def get_pub_date(self) -> Optional[str]:
        """
        Getter method corresponding to the ``prism:publicationDate`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_publication_date)

    @accepts_only('literal')
    def has_pub_date(self, string: str) -> None:
        """
        Setter method corresponding to the ``prism:publicationDate`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The date of publication of the bibliographic resource.`

        :param string: The value that will be set as the object of the property related to this method. **It must
          be a string compliant with the** ``ISO 8601`` **standard.**
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        cur_type, string = get_datatype_from_iso_8601(string)
        if cur_type is not None and string is not None:
            self.remove_pub_date()
            self._create_literal(GraphEntity.iri_has_publication_date, string, cur_type, False)

    def remove_pub_date(self) -> None:
        """
        Remover method corresponding to the ``prism:publicationDate`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_publication_date, None))

    # IS EMBODIED AS (ResourceEmbodiment)
    def get_formats(self) -> List[ResourceEmbodiment]:
        """
        Getter method corresponding to the ``frbr:embodiment`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_embodiment, 're')
        result: List[ResourceEmbodiment] = []
        for uri in uri_list:
            result.append(self.g_set.add_re(self.resp_agent, self.source, uri))
        return result

    @accepts_only('re')
    def has_format(self, re_res: ResourceEmbodiment) -> None:
        """
        Setter method corresponding to the ``frbr:embodiment`` RDF predicate.

        `The corpus identifier of the resource embodiment defining the format in which the
        bibliographic resource has been embodied, which can be either print or digital.`

        :param re_res: The value that will be set as the object of the property related to this method
        :type re_res: ResourceEmbodiment
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_embodiment, re_res.res))

    @accepts_only('re')
    def remove_format(self, re_res: ResourceEmbodiment = None) -> None:
        """
        Remover method corresponding to the ``frbr:embodiment`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param re_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type re_res: ResourceEmbodiment
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if re_res is not None:
            self.g.remove((self.res, GraphEntity.iri_embodiment, re_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_embodiment, None))

    # HAS NUMBER
    def get_number(self) -> Optional[str]:
        """
        Getter method corresponding to the ``fabio:hasSequenceIdentifier`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_sequence_identifier)

    @accepts_only('literal')
    def has_number(self, string: str) -> None:
        """
        Setter method corresponding to the ``fabio:hasSequenceIdentifier`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `A literal (for example a number or a letter) that identifies the sequence position of the
        bibliographic resource as a particular item within a larger collection (e.g. an article
        number within a journal issue, a volume number of a journal, a chapter number within
        a book).`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_number()
        self._create_literal(GraphEntity.iri_has_sequence_identifier, string)

    def remove_number(self) -> None:
        """
        Remover method corresponding to the ``fabio:hasSequenceIdentifier`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_sequence_identifier, None))

    # HAS EDITION
    def get_edition(self) -> Optional[str]:
        """
        Getter method corresponding to the ``prism:edition`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_edition)

    @accepts_only('literal')
    def has_edition(self, string: str) -> None:
        """
        Setter method corresponding to the ``prism:edition`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `An identifier for one of several alternative editions of a particular bibliographic
        resource.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_edition()
        self._create_literal(GraphEntity.iri_has_edition, string)

    def remove_edition(self) -> None:
        """
        Remover method corresponding to the ``prism:edition`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_edition, None))

    # HAS PART (BibliographicReference)
    def get_contained_in_reference_lists(self) -> List[BibliographicReference]:
        """
        Getter method corresponding to the ``frbr:part`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_contains_reference, 'be')
        result: List[BibliographicReference] = []
        for uri in uri_list:
            result.append(self.g_set.add_be(self.resp_agent, self.source, uri))
        return result

    @accepts_only('be')
    def contains_in_reference_list(self, be_res: BibliographicReference) -> None:
        """
        Setter method corresponding to the ``frbr:part`` RDF predicate.

        `A bibliographic reference within the bibliographic resource, or a discourse element
        wherein the text of the bibliographic resources can be organized.`

        :param be_res: The value that will be set as the object of the property related to this method
        :type be_res: BibliographicReference
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_contains_reference, be_res.res))

    @accepts_only('be')
    def remove_contained_in_reference_list(self, be_res: BibliographicReference = None) -> None:
        """
        Remover method corresponding to the ``frbr:part`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param be_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type be_res: BibliographicReference
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if be_res is not None:
            self.g.remove((self.res, GraphEntity.iri_contains_reference, be_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_contains_reference, None))

    # HAS PART (DiscourseElement)
    def get_contained_discourse_elements(self) -> List[DiscourseElement]:
        """
        Getter method corresponding to the ``frbr:part`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_contains_de, 'de')
        result: List[DiscourseElement] = []
        for uri in uri_list:
            result.append(self.g_set.add_de(self.resp_agent, self.source, uri))
        return result

    @accepts_only('de')
    def contains_discourse_element(self, de_res: DiscourseElement) -> None:
        """
        Setter method corresponding to the ``frbr:part`` RDF predicate.

        `A bibliographic reference within the bibliographic resource, or a discourse element
        wherein the text of the bibliographic resources can be organized.`

        :param de_res: The value that will be set as the object of the property related to this method
        :type de_res: DiscourseElement
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_contains_de, de_res.res))

    @accepts_only('de')
    def remove_contained_discourse_element(self, de_res: DiscourseElement = None) -> None:
        """
        Remover method corresponding to the ``frbr:part`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param de_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type de_res: DiscourseElement
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if de_res is not None:
            self.g.remove((self.res, GraphEntity.iri_contains_de, de_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_contains_de, None))

    # HAS CONTRIBUTOR (AgentRole)
    def get_contributors(self) -> List[AgentRole]:
        """
        Getter method corresponding to the ``pro:isDocumentContextFor`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_is_document_context_for, 'ar')
        result: List[AgentRole] = []
        for uri in uri_list:
            result.append(self.g_set.add_ar(self.resp_agent, self.source, uri))
        return result

    @accepts_only('ar')
    def has_contributor(self, ar_res: AgentRole):
        """
        Setter method corresponding to the ``pro:isDocumentContextFor`` RDF predicate.

        :param ar_res: The value that will be set as the object of the property related to this method
        :type ar_res: AgentRole
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_is_document_context_for, ar_res.res))

    @accepts_only('ar')
    def remove_contributor(self, ar_res: AgentRole = None):
        """
        Remover method corresponding to the ``frbr:part`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param ar_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type ar_res: AgentRole
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if ar_res is not None:
            self.g.remove((self.res, GraphEntity.iri_is_document_context_for, ar_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_is_document_context_for, None))

    # HAS RELATED DOCUMENT
    def get_related_documents(self) -> List[URIRef]:
        """
        Getter method corresponding to the ``dcterms:relation`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_relation)
        return uri_list

    @accepts_only('thing')
    def has_related_document(self, thing_res: URIRef) -> None:
        """
        Setter method corresponding to the ``dcterms:relation`` RDF predicate.

        `A document external to the Corpus, that is related to the bibliographic resource (such
        as a version of the bibliographic resource – for example a preprint – recorded in an
        external database).`

        :param thing_res: The value that will be set as the object of the property related to this method
        :type thing_res: URIRef
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_relation, thing_res))

    @accepts_only('thing')
    def remove_related_document(self, thing_res: URIRef = None) -> None:
        """
        Remover method corresponding to the ``dcterms:relation`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param thing_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type thing_res: URIRef
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if thing_res is not None:
            self.g.remove((self.res, GraphEntity.iri_relation, thing_res))
        else:
            self.g.remove((self.res, GraphEntity.iri_relation, None))
    
    def create_abstract(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Abstract``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_abstract)

    # HAS TYPE
    def create_archival_document(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ArchivalDocument``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_archival_document)

    def create_audio_document(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:AudioDocument``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_audio_document)

    def create_book(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Book``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_book)

    def create_book_chapter(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:BookChapter``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_book_chapter)

    def create_book_part(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Part``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_part)

    def create_book_section(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ExpressionCollection``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_expression_collection)

    def create_book_series(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:BookSeries``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_book_series)

    def create_book_set(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:BookSet``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_book_set)

    def create_book_track(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Expression``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_expression)

    def create_component(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Expression``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_expression)

    def create_computer_program(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ComputerProgram``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_computer_program)

    def create_data_management_plan(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:DataManagementPlan``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_data_management_plan)

    def create_dataset(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:DataFile``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_data_file)

    def create_dissertation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Thesis``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_thesis)

    def create_edited_book(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Book``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_book)

    def create_journal_article(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:JournalArticle``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_journal_article)

    def create_journal_editorial(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:JournalEditorial``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_journal_editorial)

    def create_issue(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:JournalIssue``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_journal_issue)

    def create_volume(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:JournalVolume``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_journal_volume)

    def create_journal(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Journal``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_journal)

    def create_monograph(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Book``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_book)

    def create_newspaper_article(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Newspaper``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_newspaper)

    def create_newspaper_editorial(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:NewspaperEditorial``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_newspaper_editorial)
    
    def create_newspaper_issue(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:NewspaperIssue``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_newspaper_issue)

    def create_peer_review(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fr:ReviewVersion``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_peer_review)

    def create_preprint(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Preprint``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_preprint)

    def create_presentation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Presentation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_presentation)

    def create_proceedings_article(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ProceedingsPaper``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_proceedings_paper)

    def create_proceedings(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:AcademicProceedings``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_academic_proceedings)

    def create_proceedings_series(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Series``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_proceedings_series)

    def create_reference_book(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ReferenceBook``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_reference_book)

    def create_reference_entry(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ReferenceEntry``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_reference_entry)

    def create_report_series(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Series``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_series)

    def create_report(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ReportDocument``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_report_document)

    def create_retraction_notice(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:RetractionNotice``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_retraction_notice)

    def create_standard_series(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Series``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_series)

    def create_standard(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:SpecificationDocument``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_specification_document)

    def create_series(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Series``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_series)

    def create_expression_collection(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:ExpressionCollection``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_expression_collection)

    def create_web_content(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:WebContent``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_web_content)

    def create_other(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``fabio:Expression``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of the bibliographic resource`

        :return: None
        """
        self._create_type(GraphEntity.iri_expression)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import json
import os
from typing import TYPE_CHECKING
from zipfile import ZipFile

import rdflib
from filelock import FileLock
from rdflib import RDF, ConjunctiveGraph, Graph, URIRef
from SPARQLWrapper import RDFXML, SPARQLWrapper

from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.support.reporter import Reporter

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Set
    from oc_ocdm.graph.graph_set import GraphSet

from pyshacl import validate


class Reader(object):

    def __init__(self, repok: Reporter = None, reperr: Reporter = None, context_map: Dict[str, Any] = None) -> None:

        if context_map is not None:
            self.context_map: Dict[str, Any] = context_map
        else:
            self.context_map: Dict[str, Any] = {}
        for context_url in self.context_map:
            ctx_file_path: Any = self.context_map[context_url]
            if type(ctx_file_path) == str and os.path.isfile(ctx_file_path):
                # This expensive operation is done only when it's really needed
                with open(ctx_file_path, 'rt', encoding='utf-8') as ctx_f:
                    self.context_map[context_url] = json.load(ctx_f)

        if repok is None:
            self.repok: Reporter = Reporter(prefix="[Reader: INFO] ")
        else:
            self.repok: Reporter = repok

        if reperr is None:
            self.reperr: Reporter = Reporter(prefix="[Reader: ERROR] ")
        else:
            self.reperr: Reporter = reperr

    def load(self, rdf_file_path: str) -> Optional[ConjunctiveGraph]:
        self.repok.new_article()
        self.reperr.new_article()

        loaded_graph: Optional[ConjunctiveGraph] = None
        if os.path.isfile(rdf_file_path):

            try:
                loaded_graph = self._load_graph(rdf_file_path)
            except Exception as e:
                self.reperr.add_sentence("[1] "
                                         "It was impossible to handle the format used for "
                                         "storing the file (stored in the temporary path) "
                                         f"'{rdf_file_path}'. Additional details: {e}")
        else:
            self.reperr.add_sentence("[2] "
                                     f"The file specified ('{rdf_file_path}') doesn't exist.")

        return loaded_graph

    def _load_graph(self, file_path: str) -> ConjunctiveGraph:
        formats: List[str] = ["json-ld", "rdfxml", "turtle", "trig", "nt11", "nquads"]

        loaded_graph: ConjunctiveGraph = ConjunctiveGraph()

        errors: str = ""
        for cur_format in formats:
            try:
                lock = FileLock(f"{file_path}.lock")
                with lock:
                    if file_path.endswith('.zip'):
                        with ZipFile(file=file_path, mode="r") as archive:
                            for zf_name in archive.namelist():
                                f = archive.open(zf_name)
                    else:
                        f = open(file_path, 'rt', encoding='utf-8')
                    if cur_format == "json-ld":
                        json_ld_file: Any = json.load(f)
                        if isinstance(json_ld_file, dict):
                            json_ld_file: List[Any] = [json_ld_file]

                        for json_ld_resource in json_ld_file:
                            # Trick to force the use of a pre-loaded context if the format
                            # specified is JSON-LD
                            if "@context" in json_ld_resource:
                                cur_context: str = json_ld_resource["@context"]
                                if cur_context in self.context_map:
                                    context_json: Any = self.context_map[cur_context]["@context"]
                                    json_ld_resource["@context"] = context_json

                            loaded_graph.parse(data=json.dumps(json_ld_resource, ensure_ascii=False),
                                            format=cur_format)
                    else:
                        loaded_graph.parse(file=f, format=cur_format)
                    f.close()
                return loaded_graph
            except Exception as e:
                errors += f" | {e}"  # Try another format

        raise IOError("1", f"It was impossible to handle the format used for storing the file '{file_path}'{errors}")

    @staticmethod
    def get_graph_from_subject(graph: Graph, subject: URIRef) -> Graph:
        g: Graph = Graph(identifier=graph.identifier)
        for p, o in graph.predicate_objects(subject):
            if isinstance(o, rdflib.term.Literal):
                o = rdflib.term.Literal(lexical_or_value=str(o), datatype=None)
            g.add((subject, p, o))
        return g

    @staticmethod
    def _extract_subjects(graph: Graph) -> Set[URIRef]:
        subjects: Set[URIRef] = set()
        for s in graph.subjects():
            subjects.add(s)
        return subjects

    def graph_validation(self, graph: Graph, closed: bool = False) -> Graph:
        valid_graph: Graph = Graph(identifier=graph.identifier)
        sg = Graph()
        if closed:
            sg.parse(os.path.join('oc_ocdm', 'resources', 'shacle_closed.ttl'))
        else:
            sg.parse(os.path.join('oc_ocdm', 'resources', 'shacle.ttl'))
        _, report_g, _ = validate(graph,
            shacl_graph=sg,
            ont_graph=None,
            inference=None,
            abort_on_first=False,
            allow_infos=False,
            allow_warnings=False,
            meta_shacl=False,
            advanced=False,
            js=False,
            debug=False)
        invalid_nodes = set()
        for triple in report_g.triples((None, URIRef('http://www.w3.org/ns/shacl#focusNode'), None)):
            invalid_nodes.add(triple[2])
        for subject in self._extract_subjects(graph):
            if subject not in invalid_nodes:
                for valid_subject_triple in graph.triples((subject, None, None)):
                    valid_graph.add(valid_subject_triple)
        return valid_graph

    @staticmethod
    def import_entities_from_graph(g_set: GraphSet, graph: Graph, resp_agent: str,
                                   enable_validation: bool = False, closed: bool = False) -> List[GraphEntity]:
        if enable_validation:
            reader = Reader()
            graph = reader.graph_validation(graph, closed)
        imported_entities: List[GraphEntity] = []
        for subject in Reader._extract_subjects(graph):
            types = []
            for o in graph.objects(subject, RDF.type):
                types.append(o)
            # ReferenceAnnotation
            if GraphEntity.iri_note in types:
                imported_entities.append(g_set.add_an(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # AgentRole
            elif GraphEntity.iri_role_in_time in types:
                imported_entities.append(g_set.add_ar(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # BibliographicReference
            elif GraphEntity.iri_bibliographic_reference in types:
                imported_entities.append(g_set.add_be(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # BibliographicResource
            elif GraphEntity.iri_expression in types:
                imported_entities.append(g_set.add_br(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # Citation
            elif GraphEntity.iri_citation in types:
                imported_entities.append(g_set.add_ci(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # DiscourseElement
            elif GraphEntity.iri_discourse_element in types:
                imported_entities.append(g_set.add_de(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # Identifier
            elif GraphEntity.iri_identifier in types:
                imported_entities.append(g_set.add_id(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # PointerList
            elif GraphEntity.iri_singleloc_pointer_list in types:
                imported_entities.append(g_set.add_pl(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # ResponsibleAgent
            elif GraphEntity.iri_agent in types:
                imported_entities.append(g_set.add_ra(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # ResourceEmbodiment
            elif GraphEntity.iri_manifestation in types:
                imported_entities.append(g_set.add_re(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
            # ReferencePointer
            elif GraphEntity.iri_intextref_pointer in types:
                imported_entities.append(g_set.add_rp(resp_agent=resp_agent, res=subject,
                                         preexisting_graph=Reader.get_graph_from_subject(graph, subject)))
        return imported_entities

    @staticmethod
    def import_entity_from_triplestore(g_set: GraphSet, ts_url: str, res: URIRef, resp_agent: str,
                                       enable_validation: bool = False) -> GraphEntity:
        sparql: SPARQLWrapper = SPARQLWrapper(ts_url)
        query: str = f"CONSTRUCT {{<{res}> ?p ?o}} WHERE {{<{res}> ?p ?o}}"
        sparql.setQuery(query)
        sparql.setMethod('GET')
        sparql.setReturnFormat(RDFXML)
        result: ConjunctiveGraph = sparql.queryAndConvert()
        if result is not None:
            imported_entities: List[GraphEntity] = Reader.import_entities_from_graph(g_set, result,
                                                                                     resp_agent, enable_validation)
            if len(imported_entities) <= 0:
                raise ValueError("The requested entity was not found or was not recognized as a proper OCDM entity.")
            else:
                return imported_entities[0]


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
import unittest

from rdflib import Namespace

from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.graph_set import GraphSet


class TestBibliographicEntity(unittest.TestCase):
    resp_agent = 'http://resp_agent.test/'

    @classmethod
    def setUpClass(cls) -> None:
        cls.graph_set = GraphSet("http://test/", "./info_dir/", "", False)

    def setUp(self):
        self.entity = self.graph_set.add_ar(self.resp_agent)
        self.identifier = self.graph_set.add_id(self.resp_agent)

    def test_has_identifier(self):
        result = self.entity.has_identifier(self.identifier)
        self.assertIsNone(result)

        triple = self.entity.res, GraphEntity.iri_has_identifier, self.identifier.res
        self.assertIn(triple, self.entity.g)

    def test_remove_duplicated_identifiers(self):
        id1 = self.graph_set.add_id(self.resp_agent)
        id2 = self.graph_set.add_id(self.resp_agent)
        id3 = self.graph_set.add_id(self.resp_agent)
        id4 = self.graph_set.add_id(self.resp_agent)

        id1.create_issn('1111-2222')
        id2.create_doi('1111-2222')
        id3.create_issn('3333-4444')
        id4.create_issn('1111-2222')

        self.entity.has_identifier(id1)
        self.entity.has_identifier(id2)
        self.entity.has_identifier(id3)
        self.entity.has_identifier(id4)

        result = self.entity.remove_duplicated_identifiers()

        self.assertIsNone(result)

        id_list = self.entity.get_identifiers()
        self.assertEqual(3, len(id_list))

        # Tuples were used down below inside the sets because they're are hashable (since they're immutable):
        id_list_set = {(i.get_scheme(), i.get_literal_value()) for i in id_list}

        datacite = Namespace("http://purl.org/spar/datacite/")
        non_duplicated_ids = {(datacite.issn, '1111-2222'),
                              (datacite.issn, '3333-4444'),
                              (datacite.doi, '1111-2222')}

        self.assertSetEqual(id_list_set, non_duplicated_ids)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import os
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from oc_ocdm.abstract_set import AbstractSet
from oc_ocdm.prov.entities.snapshot_entity import SnapshotEntity
from oc_ocdm.support.query_utils import get_update_query

if TYPE_CHECKING:
    from typing import Optional, Tuple, List, Dict, ClassVar
    from oc_ocdm.graph.graph_entity import GraphEntity

from rdflib import Graph, URIRef

from oc_ocdm.counter_handler.counter_handler import CounterHandler
from oc_ocdm.counter_handler.filesystem_counter_handler import \
    FilesystemCounterHandler
from oc_ocdm.counter_handler.in_memory_counter_handler import \
    InMemoryCounterHandler
from oc_ocdm.counter_handler.sqlite_counter_handler import SqliteCounterHandler
from oc_ocdm.graph.graph_set import GraphSet
from oc_ocdm.prov.prov_entity import ProvEntity
from oc_ocdm.support.support import (get_count, get_prefix, get_short_name,
                                     has_supplier_prefix)


class ProvSet(AbstractSet):
    # Labels
    labels: ClassVar[Dict[str, str]] = {
        "se": "snapshot of entity metadata"
    }

    def __init__(self, prov_subj_graph_set: GraphSet, base_iri: str, info_dir: str = "",
                 wanted_label: bool = True, custom_counters : dict = dict()) -> None:
        super(ProvSet, self).__init__()
        self.prov_g: GraphSet = prov_subj_graph_set
        # The following variable maps a URIRef with the related provenance entity
        self.res_to_entity: Dict[URIRef, ProvEntity] = {}
        self.base_iri: str = base_iri
        self.wanted_label: bool = wanted_label
        self.info_dir = info_dir
        short_names = ["an", "ar", "be", "br", "ci", "de", "id", "pl", "ra", "re", "rp"]
        self.counter_handlers : Dict[str, CounterHandler] = dict()
        self.custom_counters = custom_counters
        if info_dir is not None and info_dir != "":
            for short_name in short_names:
                if short_name not in custom_counters:
                    self.counter_handlers[short_name] = FilesystemCounterHandler(info_dir)
                else:
                    self.counter_handlers[short_name] = custom_counters[short_name]
        else:
            for short_name in short_names:
                if short_name not in custom_counters:
                    self.counter_handlers[short_name] = InMemoryCounterHandler()
                else:
                    self.counter_handlers[short_name] = custom_counters[short_name]        

    def get_entity(self, res: URIRef) -> Optional[ProvEntity]:
        if res in self.res_to_entity:
            return self.res_to_entity[res]

    def add_se(self, prov_subject: GraphEntity, res: URIRef = None) -> SnapshotEntity:
        if res is not None and get_short_name(res) != "se":
            raise ValueError(f"Given res: <{res}> is inappropriate for a SnapshotEntity entity.")
        if res is not None and res in self.res_to_entity:
            return self.res_to_entity[res]
        g_prov: str = str(prov_subject) + "/prov/"
        cur_g, count, label = self._add_prov(g_prov, "se", prov_subject, res)
        return SnapshotEntity(prov_subject, cur_g, self, res, prov_subject.resp_agent,
                              prov_subject.source, ProvEntity.iri_entity, count, label, "se")

    def _create_snapshot(self, cur_subj: GraphEntity, cur_time: str) -> SnapshotEntity:
        new_snapshot: SnapshotEntity = self.add_se(prov_subject=cur_subj)
        new_snapshot.is_snapshot_of(cur_subj)
        new_snapshot.has_generation_time(cur_time)
        if cur_subj.source is not None:
            new_snapshot.has_primary_source(URIRef(cur_subj.source))
        if cur_subj.resp_agent is not None:
            new_snapshot.has_resp_agent(URIRef(cur_subj.resp_agent))
        return new_snapshot

    def _get_snapshots_from_merge_list(self, cur_subj: GraphEntity) -> List[SnapshotEntity]:
        snapshots_list: List[SnapshotEntity] = []
        for entity in cur_subj.merge_list:
            last_entity_snapshot_res: Optional[URIRef] = self._retrieve_last_snapshot(entity.res)
            if last_entity_snapshot_res is not None:
                snapshots_list.append(self.add_se(prov_subject=entity, res=last_entity_snapshot_res))
        return snapshots_list

    @staticmethod
    def _get_merge_description(cur_subj: GraphEntity, snapshots_list: List[SnapshotEntity]) -> str:
        merge_description: str = f"The entity '{cur_subj.res}' has been merged"
        is_first: bool = True
        for snapshot in snapshots_list:
            if is_first:
                merge_description += f" with '{snapshot.prov_subject.res}'"
                is_first = False
            else:
                merge_description += f", '{snapshot.prov_subject.res}'"
        merge_description += "."
        return merge_description

    def generate_provenance(self, c_time: float = None) -> None:
        if c_time is None:
            cur_time: str = datetime.now(tz=timezone.utc).replace(microsecond=0).isoformat(sep="T")
        else:
            cur_time: str = datetime.fromtimestamp(c_time, tz=timezone.utc).replace(microsecond=0).isoformat(sep="T")

        # MERGED ENTITIES
        for cur_subj in self.prov_g.res_to_entity.values():
            if cur_subj is None or (not cur_subj.was_merged or cur_subj.to_be_deleted):
                # Here we must skip every entity that was not merged or that must be deleted.
                continue

            # Previous snapshot
            last_snapshot_res: Optional[URIRef] = self._retrieve_last_snapshot(cur_subj.res)
            if last_snapshot_res is None:
                # CREATION SNAPSHOT
                cur_snapshot: SnapshotEntity = self._create_snapshot(cur_subj, cur_time)
                cur_snapshot.has_description(f"The entity '{cur_subj.res}' has been created.")
            else:
                update_query: str = get_update_query(cur_subj, entity_type="graph")[0]
                was_modified: bool = (update_query != "")
                snapshots_list: List[SnapshotEntity] = self._get_snapshots_from_merge_list(cur_subj)

                if was_modified and len(snapshots_list) <= 0:
                    # MODIFICATION SNAPSHOT
                    last_snapshot: SnapshotEntity = self.add_se(prov_subject=cur_subj, res=last_snapshot_res)
                    last_snapshot.has_invalidation_time(cur_time)

                    cur_snapshot: SnapshotEntity = self._create_snapshot(cur_subj, cur_time)
                    cur_snapshot.derives_from(last_snapshot)
                    cur_snapshot.has_update_action(update_query)
                    cur_snapshot.has_description(f"The entity '{cur_subj.res}' has been modified.")
                elif len(snapshots_list) > 0:
                    # MERGE SNAPSHOT
                    last_snapshot: SnapshotEntity = self.add_se(prov_subject=cur_subj, res=last_snapshot_res)
                    last_snapshot.has_invalidation_time(cur_time)

                    cur_snapshot: SnapshotEntity = self._create_snapshot(cur_subj, cur_time)
                    cur_snapshot.derives_from(last_snapshot)
                    for snapshot in snapshots_list:
                        cur_snapshot.derives_from(snapshot)
                    if update_query:
                        cur_snapshot.has_update_action(update_query)
                    cur_snapshot.has_description(self._get_merge_description(cur_subj, snapshots_list))

        # EVERY OTHER ENTITY
        for cur_subj in self.prov_g.res_to_entity.values():
            if cur_subj is None or (cur_subj.was_merged and not cur_subj.to_be_deleted):
                # Here we must skip every entity which was merged while not being marked as to be deleted,
                # since we already processed those entities in the previous loop.
                continue

            last_snapshot_res: Optional[URIRef] = self._retrieve_last_snapshot(cur_subj.res)
            if last_snapshot_res is None:
                if cur_subj.to_be_deleted:
                    # We can ignore this entity because it was deleted even before being created.
                    pass
                else:
                    # CREATION SNAPSHOT
                    cur_snapshot: SnapshotEntity = self._create_snapshot(cur_subj, cur_time)
                    cur_snapshot.has_description(f"The entity '{cur_subj.res}' has been created.")
            else:
                update_query: str = get_update_query(cur_subj, entity_type="graph")[0]
                was_modified: bool = (update_query != "")
                if cur_subj.to_be_deleted:
                    # DELETION SNAPSHOT
                    last_snapshot: SnapshotEntity = self.add_se(prov_subject=cur_subj, res=last_snapshot_res)
                    last_snapshot.has_invalidation_time(cur_time)

                    cur_snapshot: SnapshotEntity = self._create_snapshot(cur_subj, cur_time)
                    cur_snapshot.derives_from(last_snapshot)
                    cur_snapshot.has_invalidation_time(cur_time)
                    cur_snapshot.has_description(f"The entity '{cur_subj.res}' has been deleted.")
                    cur_snapshot.has_update_action(update_query)
                elif was_modified:
                    # MODIFICATION SNAPSHOT
                    last_snapshot: SnapshotEntity = self.add_se(prov_subject=cur_subj, res=last_snapshot_res)
                    last_snapshot.has_invalidation_time(cur_time)

                    cur_snapshot: SnapshotEntity = self._create_snapshot(cur_subj, cur_time)
                    cur_snapshot.derives_from(last_snapshot)
                    cur_snapshot.has_description(f"The entity '{cur_subj.res}' has been modified.")
                    cur_snapshot.has_update_action(update_query)
    
    def _fix_info_dir(self, prov_subject: URIRef) -> None:
        short_name = get_short_name(prov_subject)
        if not short_name or self.info_dir is None or self.info_dir == "":
            return
        if not isinstance(self.counter_handlers[short_name], FilesystemCounterHandler):
            return
        if has_supplier_prefix(prov_subject, self.base_iri):
            supplier_prefix = get_prefix(prov_subject)
            info_dir_folders = os.path.normpath(self.info_dir).split(os.sep)
            info_dir_prefix = [
                folder for folder in info_dir_folders 
                if folder.startswith('0') and folder.endswith('0') and folder.isdigit() and len(folder) > 2]
            if info_dir_prefix:
                info_dir_prefix = info_dir_prefix[-1]
                if supplier_prefix != info_dir_prefix:
                    new_info_dir = os.sep.join([folder if folder != info_dir_prefix else supplier_prefix for folder in info_dir_folders])
                    self.info_dir = new_info_dir
                    self.counter_handlers[short_name]: CounterHandler = FilesystemCounterHandler(new_info_dir)

    def _add_prov(self, graph_url: str, short_name: str, prov_subject: GraphEntity,
                  res: URIRef = None) -> Tuple[Graph, Optional[str], Optional[str]]:
        self._fix_info_dir(prov_subject.res)
        cur_g: Graph = Graph(identifier=graph_url)
        self._set_ns(cur_g)

        count: Optional[str] = None
        label: Optional[str] = None

        if res is not None:
            try:
                res_count: int = int(get_count(res))
            except ValueError:
                res_count: int = -1
            if isinstance(self.counter_handlers[prov_subject.short_name], SqliteCounterHandler):
                cur_count: str = self.counter_handlers[prov_subject.short_name].read_counter(prov_subject)
            else:
                cur_count: str = self.counter_handlers[prov_subject.short_name].read_counter(prov_subject.short_name, "se", int(get_count(prov_subject.res)))
            if res_count > cur_count:
                if isinstance(self.counter_handlers[prov_subject.short_name], SqliteCounterHandler):
                    self.counter_handlers[prov_subject.short_name].set_counter(int(get_count(prov_subject.res)), prov_subject)
                else:
                    self.counter_handlers[prov_subject.short_name].set_counter(res_count, prov_subject.short_name, "se", int(get_count(prov_subject.res)))
            return cur_g, count, label
        if isinstance(self.counter_handlers[prov_subject.short_name], SqliteCounterHandler):
            count = str(self.counter_handlers[prov_subject.short_name].increment_counter(prov_subject))
        else:
            count = str(self.counter_handlers[prov_subject.short_name].increment_counter(prov_subject.short_name, "se", int(get_count(prov_subject.res))))
        if self.wanted_label:
            cur_short_name = prov_subject.short_name
            cur_entity_count = get_count(prov_subject.res)
            cur_entity_prefix = get_prefix(prov_subject.res)

            related_to_label = "related to %s %s%s" % (GraphSet.labels[cur_short_name], cur_entity_prefix,
                                                       cur_entity_count)
            related_to_short_label = "-> %s/%s%s" % (cur_short_name, cur_entity_prefix, cur_entity_count)

            label = "%s %s %s [%s/%s %s]" % (self.labels[short_name], count, related_to_label, short_name, count,
                                             related_to_short_label)

        return cur_g, count, label

    @staticmethod
    def _set_ns(g: Graph) -> None:
        g.namespace_manager.bind("prov", ProvEntity.PROV)

    def _retrieve_last_snapshot(self, prov_subject: URIRef) -> Optional[URIRef]:
        self._fix_info_dir(prov_subject)
        subj_short_name: str = get_short_name(prov_subject)
        subj_count: str = get_count(prov_subject)
        if subj_short_name not in self.custom_counters:
            try:
                if int(subj_count) <= 0:
                    raise ValueError('prov_subject is not a valid URIRef. Extracted count value should be a positive '
                                    'non-zero integer number!')
            except ValueError:
                raise ValueError('prov_subject is not a valid URIRef. Unable to extract the count value!')
        if isinstance(self.counter_handlers[subj_short_name], SqliteCounterHandler):
            last_snapshot_count: str = str(self.counter_handlers[subj_short_name].read_counter(prov_subject))
        else:
            last_snapshot_count: str = str(self.counter_handlers[subj_short_name].read_counter(subj_short_name, "se", int(subj_count)))
        if int(last_snapshot_count) <= 0:
            return None
        else:
            return URIRef(str(prov_subject) + '/prov/se/' + last_snapshot_count)

    def get_se(self) -> Tuple[SnapshotEntity]:
        result: Tuple[SnapshotEntity] = tuple()
        for ref in self.res_to_entity:
            entity: ProvEntity = self.res_to_entity[ref]
            if isinstance(entity, SnapshotEntity):
                result += (entity, )
        return result


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from rdflib import XSD

from oc_ocdm.decorators import accepts_only
from oc_ocdm.support.support import get_datatype_from_iso_8601

if TYPE_CHECKING:
    from typing import Optional
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.bibliographic_resource import BibliographicResource
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class Citation(BibliographicEntity):
    """Citation (short: ci): a permanent conceptual directional link from the citing
       bibliographic resource to a cited bibliographic resource. A citation is created by the
       performative act of an author citing a published work that is relevant to the current
       work by using a particular textual device. Typically, citations are made by including a
       bibliographic reference in the reference list of the citing work and by denoting such a
       bibliographic reference using one or more in-text reference pointers (e.g. '[1]'), or by
       the inclusion within the citing work of a link, in the form of an HTTP Uniform Resource
       Locator (URL), to the cited bibliographic resource on the World Wide Web."""

    @accepts_only('ci')
    def merge(self, other: Citation) -> None:
        """
        The merge operation allows combining two ``Citation`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``Citation``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: Citation
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(Citation, self).merge(other)

        citing_br: Optional[BibliographicResource] = other.get_citing_entity()
        if citing_br is not None:
            self.has_citing_entity(citing_br)

        cited_br: Optional[BibliographicResource] = other.get_cited_entity()
        if cited_br is not None:
            self.has_cited_entity(cited_br)

        creation_date: Optional[str] = other.get_citation_creation_date()
        if creation_date is not None:
            self.has_citation_creation_date(creation_date)

        time_span: Optional[str] = other.get_citation_time_span()
        if time_span is not None:
            self.has_citation_time_span(time_span)

        characterization: Optional[URIRef] = other.get_citation_characterization()
        if characterization is not None:
            self.has_citation_characterization(characterization)

    # HAS CITING DOCUMENT (BibliographicResource)
    def get_citing_entity(self) -> Optional[BibliographicResource]:
        """
        Getter method corresponding to the ``cito:hasCitingEntity`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_has_citing_entity, 'br')
        if uri is not None:
            return self.g_set.add_br(self.resp_agent, self.source, uri)

    @accepts_only('br')
    def has_citing_entity(self, citing_res: BibliographicResource) -> None:
        """
        Setter method corresponding to the ``cito:hasCitingEntity`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The bibliographic resource which acts as the source for the citation.`

        :param citing_res: The value that will be set as the object of the property related to this method
        :type citing_res: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_citing_entity()
        self.g.add((self.res, GraphEntity.iri_has_citing_entity, citing_res.res))

    def remove_citing_entity(self) -> None:
        """
        Remover method corresponding to the ``cito:hasCitingEntity`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_citing_entity, None))

    # HAS CITED DOCUMENT (BibliographicResource)
    def get_cited_entity(self) -> Optional[BibliographicResource]:
        """
        Getter method corresponding to the ``cito:hasCitedEntity`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_has_cited_entity, 'br')
        if uri is not None:
            return self.g_set.add_br(self.resp_agent, self.source, uri)

    @accepts_only('br')
    def has_cited_entity(self, cited_res: BibliographicResource) -> None:
        """
        Setter method corresponding to the ``cito:hasCitedEntity`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The bibliographic resource which acts as the target for the citation.`

        :param cited_res: The value that will be set as the object of the property related to this method
        :type cited_res: BibliographicResource
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_cited_entity()
        self.g.add((self.res, GraphEntity.iri_has_cited_entity, cited_res.res))

    def remove_cited_entity(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasCitedEntity`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_cited_entity, None))

    # HAS CITATION CREATION DATE
    def get_citation_creation_date(self) -> Optional[str]:
        """
        Getter method corresponding to the ``cito:hasCitationCreationDate`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_citation_creation_date)

    @accepts_only('literal')
    def has_citation_creation_date(self, string: str) -> None:
        """
        Setter method corresponding to the ``cito:hasCitationCreationDate`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The date on which the citation was created. This has the same numerical value
        as the publication date of the citing bibliographic resource, but is a property
        of the citation itself. When combined with the citation time span, it permits
        that citation to be located in history.`

        :param string: The value that will be set as the object of the property related to this method. **It must
          be a string compliant with the** ``ISO 8601`` **standard.**
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        cur_type, string = get_datatype_from_iso_8601(string)
        if cur_type is not None and string is not None:
            self.remove_citation_creation_date()
            self._create_literal(GraphEntity.iri_has_citation_creation_date, string, cur_type, False)

    def remove_citation_creation_date(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasCitationCreationDate`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_citation_creation_date, None))

    # HAS CITATION TIME SPAN
    def get_citation_time_span(self) -> Optional[str]:
        """
        Getter method corresponding to the ``cito:hasCitationTimeSpan`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_citation_time_span)

    @accepts_only('literal')
    def has_citation_time_span(self, string: str) -> None:
        """
        Setter method corresponding to the ``cito:hasCitationTimeSpan`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The date interval between the publication date of the cited bibliographic resource and
        the publication date of the citing bibliographic resource.`

        :param string: The value that will be set as the object of the property related to this method. **It must
          be a string compliant with the** ``xsd:duration`` **datatype.**
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_citation_time_span()
        self._create_literal(GraphEntity.iri_has_citation_time_span, string, XSD.duration, False)

    def remove_citation_time_span(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasCitationTimeSpan`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_citation_time_span, None))

    # HAS CITATION CHARACTERIZATION
    def get_citation_characterization(self) -> Optional[URIRef]:
        """
        Getter method corresponding to the ``cito:hasCitationCharacterisation`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_citation_characterisation)
        return uri

    @accepts_only('thing')
    def has_citation_characterization(self, thing_res: URIRef) -> None:
        """
        Setter method corresponding to the ``cito:hasCitationCharacterisation`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The citation function characterizing the purpose of the citation.`

        :param thing_res: The value that will be set as the object of the property related to this method
        :type thing_res: URIRef
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_citation_characterization()
        self.g.add((self.res, GraphEntity.iri_citation_characterisation, thing_res))

    def remove_citation_characterization(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasCitationCharacterisation`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_citation_characterisation, None))

    # HAS TYPE
    def create_self_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:SelfCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_self_citation)

    def create_affiliation_self_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:AffiliationSelfCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_affiliation_self_citation)

    def create_author_network_self_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:AuthorNetworkSelfCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_author_network_self_citation)

    def create_author_self_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:AuthorSelfCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_author_self_citation)

    def create_funder_self_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:FunderSelfCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_funder_self_citation)

    def create_journal_self_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:JournalSelfCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_journal_self_citation)

    def create_journal_cartel_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:JournalCartelCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_journal_cartel_citation)

    def create_distant_citation(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``cito:DistantCitation``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        :return: None
        """
        self._create_type(GraphEntity.iri_distant_citation)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
import unittest

from rdflib import URIRef, Literal, XSD

from oc_ocdm.graph.graph_set import GraphSet
from oc_ocdm.prov.prov_entity import ProvEntity
from oc_ocdm.prov.prov_set import ProvSet


class TestSnapshotEntity(unittest.TestCase):
    resp_agent = 'http://resp_agent.test/'

    @classmethod
    def setUpClass(cls) -> None:
        cls.graph_set = GraphSet("http://test/", "./info_dir/", "", False)
        cls.prov_set = ProvSet(cls.graph_set, "http://test/", "./info_dir/", False)

    def setUp(self):
        self.prov_subject = self.graph_set.add_br(self.resp_agent)
        self.se = self.prov_set.add_se(self.prov_subject)
        self.prev_se = self.prov_set.add_se(self.prov_subject)

    def test_has_generation_time(self):
        time = "2001-10-26T21:32:52"
        datatype = XSD.dateTime
        result = self.se.has_generation_time(time)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_generated_at_time, Literal(time, datatype=datatype,
                                                                        normalize=False)
        self.assertIn(triple, self.se.g)

    def test_has_invalidation_time(self):
        time = "2001-10-26T21:32:52"
        datatype = XSD.dateTime
        result = self.se.has_invalidation_time(time)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_invalidated_at_time, Literal(time, datatype=datatype,
                                                                          normalize=False)
        self.assertIn(triple, self.se.g)

    def test_is_snapshot_of(self):
        ar = self.graph_set.add_ar(self.resp_agent)
        result = self.se.is_snapshot_of(self.prov_subject)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_specialization_of, self.prov_subject.res
        self.assertIn(triple, self.se.g)

    def test_derives_from(self):
        result = self.se.derives_from(self.prev_se)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_was_derived_from, self.prev_se.res
        self.assertIn(triple, self.se.g)

    def test_has_primary_source(self):
        primary_source = URIRef("http://test/primarySource")
        result = self.se.has_primary_source(primary_source)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_had_primary_source, primary_source
        self.assertIn(triple, self.se.g)

    def test_has_update_action(self):
        update_query = "DELETE {} INSERT {}"
        result = self.se.has_update_action(update_query)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_has_update_query, Literal(update_query)
        self.assertIn(triple, self.se.g)

    def test_has_description(self):
        description = "Description"
        result = self.se.has_description(description)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_description, Literal(description)
        self.assertIn(triple, self.se.g)

    def test_has_resp_agent(self):
        ra = URIRef("http://test/ra")
        result = self.se.has_resp_agent(ra)
        self.assertIsNone(result)

        triple = self.se.res, ProvEntity.iri_was_attributed_to, ra
        self.assertIn(triple, self.se.g)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

import os
from shutil import copymode, move
from tempfile import mkstemp
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import BinaryIO, Tuple, List, Dict

from oc_ocdm.counter_handler.counter_handler import CounterHandler
from oc_ocdm.support.support import is_string_empty


class FilesystemCounterHandler(CounterHandler):
    """A concrete implementation of the ``CounterHandler`` interface that persistently stores
    the counter values within the filesystem."""

    _initial_line_len: int = 3
    _trailing_char: str = " "

    def __init__(self, info_dir: str) -> None:
        """
        Constructor of the ``FilesystemCounterHandler`` class.

        :param info_dir: The path to the folder that does/will contain the counter values.
        :type info_dir: str
        :raises ValueError: if ``info_dir`` is None or an empty string.
        """
        if info_dir is None or is_string_empty(info_dir):
            raise ValueError("info_dir parameter is required!")

        if info_dir[-1] != os.sep:
            info_dir += os.sep

        self.info_dir: str = info_dir
        self.datasets_dir: str = info_dir + 'datasets' + os.sep
        self.short_names: List[str] = ["an", "ar", "be", "br", "ci", "de", "id", "pl", "ra", "re", "rp"]
        self.metadata_short_names: List[str] = ["di"]
        self.info_files: Dict[str, str] = {key: ("info_file_" + key + ".txt")
                                           for key in self.short_names}
        self.prov_files: Dict[str, str] = {key: ("prov_file_" + key + ".txt")
                                           for key in self.short_names}

    def set_counter(self, new_value: int, entity_short_name: str, prov_short_name: str = "",
                    identifier: int = 1) -> None:
        """
        It allows to set the counter value of graph and provenance entities.

        :param new_value: The new counter value to be set
        :type new_value: int
        :param entity_short_name: The short name associated either to the type of the entity itself
         or, in case of a provenance entity, to the type of the relative graph entity.
        :type entity_short_name: str
        :param prov_short_name: In case of a provenance entity, the short name associated to the type
         of the entity itself. An empty string otherwise.
        :type prov_short_name: str
        :param identifier: In case of a provenance entity, the counter value that identifies the relative
          graph entity. The integer value '1' otherwise.
        :type identifier: int
        :raises ValueError: if ``new_value`` is a negative integer or ``identifier`` is less than or equal to zero.
        :return: None
        """
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")

        if prov_short_name == "se":
            file_path: str = self._get_prov_path(entity_short_name)
        else:
            file_path: str = self._get_info_path(entity_short_name)
        self._set_number(new_value, file_path, identifier)

    def read_counter(self, entity_short_name: str, prov_short_name: str = "", identifier: int = 1) -> int:
        """
        It allows to read the counter value of graph and provenance entities.

        :param entity_short_name: The short name associated either to the type of the entity itself
         or, in case of a provenance entity, to the type of the relative graph entity.
        :type entity_short_name: str
        :param prov_short_name: In case of a provenance entity, the short name associated to the type
         of the entity itself. An empty string otherwise.
        :type prov_short_name: str
        :param identifier: In case of a provenance entity, the counter value that identifies the relative
          graph entity. The integer value '1' otherwise.
        :type identifier: int
        :raises ValueError: if ``identifier`` is less than or equal to zero.
        :return: The requested counter value.
        """
        if prov_short_name == "se":
            file_path: str = self._get_prov_path(entity_short_name)
        else:
            file_path: str = self._get_info_path(entity_short_name)
        return self._read_number(file_path, identifier)[0]

    def increment_counter(self, entity_short_name: str, prov_short_name: str = "", identifier: int = 1) -> int:
        """
        It allows to increment the counter value of graph and provenance entities by one unit.

        :param entity_short_name: The short name associated either to the type of the entity itself
         or, in case of a provenance entity, to the type of the relative graph entity.
        :type entity_short_name: str
        :param prov_short_name: In case of a provenance entity, the short name associated to the type
         of the entity itself. An empty string otherwise.
        :type prov_short_name: str
        :param identifier: In case of a provenance entity, the counter value that identifies the relative
          graph entity. The integer value '1' otherwise.
        :type identifier: int
        :raises ValueError: if ``identifier`` is less than or equal to zero.
        :return: The newly-updated (already incremented) counter value.
        """
        if prov_short_name == "se":
            file_path: str = self._get_prov_path(entity_short_name)
        else:
            file_path: str = self._get_info_path(entity_short_name)
        return self._add_number(file_path, identifier)

    def _get_info_path(self, short_name: str) -> str:
        return self.info_dir + self.info_files[short_name]

    def _get_prov_path(self, short_name: str) -> str:
        return self.info_dir + self.prov_files[short_name]

    def _get_metadata_path(self, short_name: str, dataset_name: str) -> str:
        return self.datasets_dir + dataset_name + os.sep + 'metadata_' + short_name + '.txt'

    def __initialize_file_if_not_existing(self, file_path: str):
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))

        if not os.path.isfile(file_path):
            with open(file_path, 'wb') as file:
                first_line: str = self._trailing_char * (self._initial_line_len - 1) + '\n'
                file.write(first_line.encode('ascii'))

    def _read_number(self, file_path: str, line_number: int) -> Tuple[int, int]:
        if line_number <= 0:
            raise ValueError("line_number must be a positive non-zero integer number!")

        self.__initialize_file_if_not_existing(file_path)

        cur_number: int = 0
        cur_line_len: int = 0
        try:
            with open(file_path, 'rb') as file:
                cur_line_len = self._get_line_len(file)
                line_offset = (line_number - 1) * cur_line_len
                file.seek(line_offset)
                line = file.readline(cur_line_len).decode('ascii')
                cur_number = int(line.rstrip(self._trailing_char + '\n'))
        except ValueError:
            cur_number = 0
        except Exception as e:
            print(e)

        return cur_number, cur_line_len

    def _add_number(self, file_path: str, line_number: int = 1) -> int:
        if line_number <= 0:
            raise ValueError("line_number must be a positive non-zero integer number!")

        self.__initialize_file_if_not_existing(file_path)

        cur_number, cur_line_len = self._read_number(file_path, line_number)
        cur_number += 1

        cur_number_len: int = len(str(cur_number)) + 1
        if cur_number_len > cur_line_len:
            self._increase_line_len(file_path, new_length=cur_number_len)
            cur_line_len = cur_number_len

        with open(file_path, 'r+b') as file:
            line_offset: int = (line_number - 1) * cur_line_len
            file.seek(line_offset)
            line: str = str(cur_number).ljust(cur_line_len - 1, self._trailing_char) + '\n'
            file.write(line.encode('ascii'))
            file.seek(-cur_line_len, os.SEEK_CUR)
            self._fix_previous_lines(file, cur_line_len)
        return cur_number

    def _set_number(self, new_value: int, file_path: str, line_number: int = 1) -> None:
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")

        if line_number <= 0:
            raise ValueError("line_number must be a positive non-zero integer number!")

        self.__initialize_file_if_not_existing(file_path)

        cur_line_len = self._read_number(file_path, line_number)[1]

        cur_number_len: int = len(str(new_value)) + 1
        if cur_number_len > cur_line_len:
            self._increase_line_len(file_path, new_length=cur_number_len)
            cur_line_len = cur_number_len

        with open(file_path, 'r+b') as file:
            line_offset: int = (line_number - 1) * cur_line_len
            file.seek(line_offset)
            line: str = str(new_value).ljust(cur_line_len - 1, self._trailing_char) + '\n'
            file.write(line.encode('ascii'))
            file.seek(-cur_line_len, os.SEEK_CUR)
            self._fix_previous_lines(file, cur_line_len)

    @staticmethod
    def _get_line_len(file: BinaryIO) -> int:
        cur_char: str = file.read(1).decode("ascii")
        count: int = 1
        while cur_char is not None and len(cur_char) == 1 and cur_char != "\0":
            cur_char = file.read(1).decode("ascii")
            count += 1
            if cur_char == "\n":
                break

        # Undo I/O pointer updates
        file.seek(0)

        if cur_char is None:
            raise EOFError("Reached end-of-file without encountering a line separator!")
        elif cur_char == "\0":
            raise ValueError("Encountered a NULL byte!")
        else:
            return count

    def _increase_line_len(self, file_path: str, new_length: int = 0) -> None:
        if new_length <= 0:
            raise ValueError("new_length must be a positive non-zero integer number!")

        with open(file_path, 'rb') as cur_file:
            if self._get_line_len(cur_file) >= new_length:
                raise ValueError("Current line length is greater than new_length!")

        fh, abs_path = mkstemp()
        with os.fdopen(fh, 'wb') as new_file:
            with open(file_path, 'rt', encoding='ascii') as old_file:
                for line in old_file:
                    number: str = line.rstrip(self._trailing_char + '\n')
                    new_line: str = str(number).ljust(new_length - 1, self._trailing_char) + '\n'
                    new_file.write(new_line.encode('ascii'))

        # Copy the file permissions from the old file to the new file
        copymode(file_path, abs_path)

        # Replace original file
        os.remove(file_path)
        move(abs_path, file_path)

    @staticmethod
    def _is_a_valid_line(buf: bytes) -> bool:
        string: str = buf.decode("ascii")
        return (string[-1] == "\n") and ("\0" not in string[:-1])

    def _fix_previous_lines(self, file: BinaryIO, line_len: int) -> None:
        if line_len < self._initial_line_len:
            raise ValueError("line_len should be at least %d!" % self._initial_line_len)

        while file.tell() >= line_len:
            file.seek(-line_len, os.SEEK_CUR)
            buf: bytes = file.read(line_len)
            if self._is_a_valid_line(buf) or len(buf) < line_len:
                break
            else:
                file.seek(-line_len, os.SEEK_CUR)
                fixed_line: str = (self._trailing_char * (line_len - 1)) + "\n"
                file.write(fixed_line.encode("ascii"))
                file.seek(-line_len, os.SEEK_CUR)

    def set_metadata_counter(self, new_value: int, entity_short_name: str, dataset_name: str) -> None:
        """
        It allows to set the counter value of metadata entities.

        :param new_value: The new counter value to be set
        :type new_value: int
        :param entity_short_name: The short name associated either to the type of the entity itself.
        :type entity_short_name: str
        :param dataset_name: In case of a ``Dataset``, its name. Otherwise, the name of the relative dataset.
        :type dataset_name: str
        :raises ValueError: if ``new_value`` is a negative integer, ``dataset_name`` is None or
          ``entity_short_name`` is not a known metadata short name.
        :return: None
        """
        if new_value < 0:
            raise ValueError("new_value must be a non negative integer!")

        if dataset_name is None:
            raise ValueError("dataset_name must be provided!")

        if entity_short_name not in self.metadata_short_names:
            raise ValueError("entity_short_name is not a known metadata short name!")

        file_path: str = self._get_metadata_path(entity_short_name, dataset_name)
        return self._set_number(new_value, file_path, 1)

    def read_metadata_counter(self, entity_short_name: str, dataset_name: str) -> int:
        """
        It allows to read the counter value of metadata entities.

        :param entity_short_name: The short name associated either to the type of the entity itself.
        :type entity_short_name: str
        :param dataset_name: In case of a ``Dataset``, its name. Otherwise, the name of the relative dataset.
        :type dataset_name: str
        :raises ValueError: if ``dataset_name`` is None or ``entity_short_name`` is not a known metadata short name.
        :return: The requested counter value.
        """
        if dataset_name is None:
            raise ValueError("dataset_name must be provided!")

        if entity_short_name not in self.metadata_short_names:
            raise ValueError("entity_short_name is not a known metadata short name!")

        file_path: str = self._get_metadata_path(entity_short_name, dataset_name)
        return self._read_number(file_path, 1)[0]

    def increment_metadata_counter(self, entity_short_name: str, dataset_name: str) -> int:
        """
        It allows to increment the counter value of metadata entities by one unit.

        :param entity_short_name: The short name associated either to the type of the entity itself.
        :type entity_short_name: str
        :param dataset_name: In case of a ``Dataset``, its name. Otherwise, the name of the relative dataset.
        :type dataset_name: str
        :raises ValueError: if ``dataset_name`` is None or ``entity_short_name`` is not a known metadata short name.
        :return: The newly-updated (already incremented) counter value.
        """
        if dataset_name is None:
            raise ValueError("dataset_name must be provided!")

        if entity_short_name not in self.metadata_short_names:
            raise ValueError("entity_short_name is not a known metadata short name!")

        file_path: str = self._get_metadata_path(entity_short_name, dataset_name)
        return self._add_number(file_path, 1)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2016, Silvio Peroni <essepuntato@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any purpose
# with or without fee is hereby granted, provided that the above copyright notice
# and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT,
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
# ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
# SOFTWARE.
from __future__ import annotations

from typing import TYPE_CHECKING

from oc_ocdm.decorators import accepts_only

if TYPE_CHECKING:
    from typing import Optional, List
    from rdflib import URIRef
    from oc_ocdm.graph.entities.bibliographic.reference_pointer import ReferencePointer
    from oc_ocdm.graph.entities.bibliographic.pointer_list import PointerList
from oc_ocdm.graph.graph_entity import GraphEntity
from oc_ocdm.graph.entities.bibliographic_entity import BibliographicEntity


class DiscourseElement(BibliographicEntity):
    """Discourse element (short: de): a document component, either structural (e.g.
       paragraph, section, chapter, table, caption, footnote, title) or rhetorical (e.g.
       introduction, discussion, acknowledgements, reference list, figure, appendix), in which
       the content of a bibliographic resource can be organized."""

    @accepts_only('de')
    def merge(self, other: DiscourseElement) -> None:
        """
        The merge operation allows combining two ``DiscourseElement`` entities into a single one,
        by marking the second entity as to be deleted while also copying its data into the current
        ``DiscourseElement``. Moreover, every triple from the containing ``GraphSet`` referring to the second
        entity gets "redirected" to the current entity: **every other reference contained inside a
        different source (e.g. a triplestore) must be manually handled by the user!**

        In case of functional properties, values from the current entity get overwritten
        by those coming from the second entity while, in all other cases, values from the
        second entity are simply appended to those of the current entity. In this context,
        ``rdfs:label`` is considered as a functional property, while ``rdf:type`` is not.

        :param other: The entity which will be marked as to be deleted and whose properties will
         be merged into the current entity.
        :type other: DiscourseElement
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        super(DiscourseElement, self).merge(other)

        title: Optional[str] = other.get_title()
        if title is not None:
            self.has_title(title)

        de_list: List[DiscourseElement] = other.get_contained_discourse_elements()
        for cur_de in de_list:
            self.contains_discourse_element(cur_de)

        next_de: Optional[DiscourseElement] = other.get_next_de()
        if next_de is not None:
            self.has_next_de(next_de)

        rp_list: List[ReferencePointer] = other.get_is_context_of_rp()
        for cur_rp in rp_list:
            self.is_context_of_rp(cur_rp)

        pl_list: List[PointerList] = other.get_is_context_of_pl()
        for cur_pl in pl_list:
            self.is_context_of_pl(cur_pl)

        content: Optional[str] = other.get_content()
        if content is not None:
            self.has_content(content)

        number: Optional[str] = other.get_number()
        if number is not None:
            self.has_number(number)

    # HAS TITLE
    def get_title(self) -> Optional[str]:
        """
        Getter method corresponding to the ``dcterms:title`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_title)

    @accepts_only('literal')
    def has_title(self, string: str) -> None:
        """
        Setter method corresponding to the ``dcterms:title`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The title of the discourse element, such as the title of a figure or a section in an article.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_title()
        self._create_literal(GraphEntity.iri_title, string)

    def remove_title(self) -> None:
        """
        Remover method corresponding to the ``dcterms:title`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_title, None))
    
    # HAS PART (DiscourseElement)
    def get_contained_discourse_elements(self) -> List[DiscourseElement]:
        """
        Getter method corresponding to the ``frbr:part`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_contains_de, 'de')
        result: List[DiscourseElement] = []
        for uri in uri_list:
            result.append(self.g_set.add_de(self.resp_agent, self.source, uri))
        return result

    @accepts_only('de')
    def contains_discourse_element(self, de_res: DiscourseElement) -> None:
        """
        Setter method corresponding to the ``frbr:part`` RDF predicate.

        `The discourse element hierarchically nested within the parent element, such as a
        sentence within a paragraph, or a paragraph within a section.`

        :param de_res: The value that will be set as the object of the property related to this method
        :type de_res: DiscourseElement
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_contains_de, de_res.res))

    @accepts_only('de')
    def remove_contained_discourse_element(self, de_res: DiscourseElement = None) -> None:
        """
        Remover method corresponding to the ``frbr:part`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param de_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type de_res: DiscourseElement
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if de_res is not None:
            self.g.remove((self.res, GraphEntity.iri_contains_de, de_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_contains_de, None))

    # HAS NEXT (DiscourseElement)
    def get_next_de(self) -> Optional[DiscourseElement]:
        """
        Getter method corresponding to the ``oco:hasNext`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        uri: Optional[URIRef] = self._get_uri_reference(GraphEntity.iri_has_next, 'de')
        if uri is not None:
            return self.g_set.add_de(self.resp_agent, self.source, uri)

    @accepts_only('de')
    def has_next_de(self, de_res: DiscourseElement) -> None:
        """
        Setter method corresponding to the ``oco:hasNext`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The following discourse element that includes at least one in-text reference pointer.`

        :param de_res: The value that will be set as the object of the property related to this method
        :type de_res: DiscourseElement
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_next_de()
        self.g.add((self.res, GraphEntity.iri_has_next, de_res.res))

    def remove_next_de(self) -> None:
        """
        Remover method corresponding to the ``oco:hasNext`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_next, None))

    # IS CONTEXT OF (ReferencePointer)
    def get_is_context_of_rp(self) -> List[ReferencePointer]:
        """
        Getter method corresponding to the ``c4o:isContextOf`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_is_context_of, 'rp')
        result: List[ReferencePointer] = []
        for uri in uri_list:
            result.append(self.g_set.add_rp(self.resp_agent, self.source, uri))
        return result

    @accepts_only('rp')
    def is_context_of_rp(self, rp_res: ReferencePointer) -> None:
        """
        Setter method corresponding to the ``c4o:isContextOf`` RDF predicate.

        `Provides the textual and semantic context of the in-text reference pointer
        that appears within the discourse element.`

        :param rp_res: The value that will be set as the object of the property related to this method
        :type rp_res: ReferencePointer
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_is_context_of, rp_res.res))

    @accepts_only('rp')
    def remove_is_context_of_rp(self, rp_res: ReferencePointer = None) -> None:
        """
        Remover method corresponding to the ``c4o:isContextOf`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param rp_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type rp_res: ReferencePointer
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if rp_res is not None:
            self.g.remove((self.res, GraphEntity.iri_is_context_of, rp_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_is_context_of, None))

    # IS CONTEXT OF (PointerList)
    def get_is_context_of_pl(self) -> List[PointerList]:
        """
        Getter method corresponding to the ``c4o:isContextOf`` RDF predicate.

        :return: A list containing the requested values if found, None otherwise
        """
        uri_list: List[URIRef] = self._get_multiple_uri_references(GraphEntity.iri_is_context_of, 'pl')
        result: List[PointerList] = []
        for uri in uri_list:
            result.append(self.g_set.add_pl(self.resp_agent, self.source, uri))
        return result

    @accepts_only('pl')
    def is_context_of_pl(self, pl_res: PointerList) -> None:
        """
        Setter method corresponding to the ``c4o:isContextOf`` RDF predicate.

        `Provides the textual and semantic context of the list of
        in-text reference pointers that appears within the discourse element.`

        :param pl_res: The value that will be set as the object of the property related to this method
        :type pl_res: PointerList
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.g.add((self.res, GraphEntity.iri_is_context_of, pl_res.res))

    @accepts_only('pl')
    def remove_is_context_of_pl(self, pl_res: PointerList = None) -> None:
        """
        Remover method corresponding to the ``c4o:isContextOf`` RDF predicate.

        **WARNING: this is a non-functional property, hence, if the parameter
        is None, any existing value will be removed!**

        :param pl_res: If not None, the specific object value that will be removed from the property
         related to this method (defaults to None)
        :type pl_res: PointerList
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if pl_res is not None:
            self.g.remove((self.res, GraphEntity.iri_is_context_of, pl_res.res))
        else:
            self.g.remove((self.res, GraphEntity.iri_is_context_of, None))

    # HAS CONTENT
    def get_content(self) -> Optional[str]:
        """
        Getter method corresponding to the ``c4o:hasContent`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_content)

    @accepts_only('literal')
    def has_content(self, string: str) -> None:
        """
        Setter method corresponding to the ``c4o:hasContent`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        `The literal document text contained by the discourse element.`

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_content()
        self._create_literal(GraphEntity.iri_has_content, string)

    def remove_content(self) -> None:
        """
        Remover method corresponding to the ``c4o:hasContent`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_content, None))

    # HAS NUMBER
    def get_number(self) -> Optional[str]:
        """
        Getter method corresponding to the ``fabio:hasSequenceIdentifier`` RDF predicate.

        :return: The requested value if found, None otherwise
        """
        return self._get_literal(GraphEntity.iri_has_sequence_identifier)

    @accepts_only('literal')
    def has_number(self, string: str) -> None:
        """
        Setter method corresponding to the ``fabio:hasSequenceIdentifier`` RDF predicate.

        **WARNING: this is a functional property, hence any existing value will be overwritten!**

        :param string: The value that will be set as the object of the property related to this method
        :type string: str
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        self.remove_number()
        self._create_literal(GraphEntity.iri_has_sequence_identifier, string)

    def remove_number(self) -> None:
        """
        Remover method corresponding to the ``fabio:hasSequenceIdentifier`` RDF predicate.

        :return: None
        """
        self.g.remove((self.res, GraphEntity.iri_has_sequence_identifier, None))

    # HAS TYPE
    @accepts_only('thing')
    def create_discourse_element(self, de_class: URIRef = None) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        If parameter is None, it implicitly sets the object value ``deo:DiscourseElement``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :param de_class: The value that will be set as the object of the property related to this method
        :type de_class: URIRef
        :raises TypeError: if the parameter is of the wrong type
        :return: None
        """
        if de_class is not None:
            self._create_type(de_class)
        else:
            self._create_type(GraphEntity.iri_discourse_element)

    def create_section(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Section``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_section)

    def create_section_title(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:SectionTitle``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_section_title)

    def create_paragraph(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Paragraph``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_paragraph)

    def create_sentence(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Sentence``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_sentence)

    def create_text_chunk(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:TextChunk``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_text_chunk)

    def create_table(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Table``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_table)

    def create_footnote(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``doco:Footnote``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_footnote)

    def create_caption(self) -> None:
        """
        Setter method corresponding to the ``rdf:type`` RDF predicate.
        It implicitly sets the object value ``deo:Caption``.

        **WARNING: the OCDM specification admits at most two types for an entity.
        The main type cannot be edited or removed. Any existing secondary type
        will be overwritten!**

        `The type of discourse element – such as “paragraph”, “section”, “sentence”,
        “acknowledgements”, “reference list” or “figure”.`

        :return: None
        """
        self._create_type(GraphEntity.iri_caption)
